/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/*===========================================================================
 * 
 * Project : freyja
 * Author  : Terry 'Mongoose' Hendrix II
 * Website : http://www.icculus.org/freyja
 * Email   : mongooseichiban@gmail.com
 * Object  : Scene
 * License : No use w/o permission (C) 2005-2007 Mongoose
 * Comments: This is the scene graph.
 * 
 *-- Test Defines -----------------------------------------------
 *           
 * UNIT_TEST_SCENE - Builds class as a console unit test. 
 *
 *-- History ------------------------------------------------ 
 *
 * 2007.12.06:
 * Mongoose - Created
 ==========================================================================*/

#ifndef GUARD__FREYJA_SCENE_H_
#define GUARD__FREYJA_SCENE_H_

#if TINYXML_FOUND
#   include <tinyxml/tinyxml.h>
#endif

#include <hel/math.h>
#include <hel/Vec3.h>
#include <hel/Mat44.h>
#include <hel/Quat.h>
#include <mstl/Vector.h>
#include <mstl/String.h>

#include "freyja.h"


namespace freyja {

class SceneObject
{
public:

	typedef enum {
		fNone        =  0,
		fHighlighted =  1,   /* This object should be rendered with a highlight.*/
		fMuted       =  2,   /* Can't edit muted objects. */
		fSelected    =  4,   /* This object is marked as selected. */
		fHidden      =  8,   /* This object shouldn't be rendered or hit. */
		fRayHit      = 16,   /* This object was marked in a hit test. */
		fLocked      = 32,   /* This object can't be accessed. */
		fReserved64  = 64    /* Currently not used. */    
	} Flags;

	const byte &GetFlags() { return mFlags; }
	/*------------------------------------------------------
	 * Pre  : 
	 * Post : Returns option flag bitmap.
	 *
	 ------------------------------------------------------*/

	void ClearFlag(Flags flag) { if (!(mFlags & fMuted)) mFlags &= ~flag; }
	/*------------------------------------------------------
	 * Pre  : 
	 * Post : Clears option flag if not mute.
	 *
	 ------------------------------------------------------*/

	void SetFlag(Flags flag) { if (!(mFlags & fMuted)) mFlags |= flag; }
	/*------------------------------------------------------
	 * Pre  : 
	 * Post : Sets option flag if not mute.
	 *
	 ------------------------------------------------------*/

	bool IsMuted() { return (mFlags & fMuted); }
	/*------------------------------------------------------
	 * Pre  : 
	 * Post : 'Unlocks' current option flags.
	 *
	 ------------------------------------------------------*/

	void ClearMuted() { mFlags &= ~fMuted; }
	/*------------------------------------------------------
	 * Pre  : 
	 * Post : 'Unlocks' current option flags.
	 *
	 ------------------------------------------------------*/

	virtual void UpdateOwnerOnDelete();

	virtual void UpdateChildrenOnDelete();

protected:

	byte mFlags;                /* State flags */

	SceneObject* mParent;

	mstl::Vector<SceneObject*> mChildren;
};


class Scene 
{
public:

	////////////////////////////////////////////////////////////
	// Constructors
	////////////////////////////////////////////////////////////

	Scene();
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : 
	 ------------------------------------------------------*/

	~Scene();
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : 
	 ------------------------------------------------------*/


	////////////////////////////////////////////////////////////
	// Gobal pool API
	////////////////////////////////////////////////////////////

	static uint32 GetCount();
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : Returns number of bones in gobal store
	 ------------------------------------------------------*/

	static Bone *GetBone(index_t uid);
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : Returns bone matching gobal store UID
	 ------------------------------------------------------*/

	static void ResetPool();
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : Erase the gobal bone store
	 ------------------------------------------------------*/

	index_t AddToPool();
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : Add this bone to the gobal store
	 *        Return UID
	 ------------------------------------------------------*/

	index_t GetUID();
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : Returns this bone's gobal store UID
	 ------------------------------------------------------*/

	void RemoveFromPool();
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : Remove this bone from gobal store
	 ------------------------------------------------------*/

#if TINYXML_FOUND
	////////////////////////////////////////////////////////////
	// Optional XML serialization.
	////////////////////////////////////////////////////////////

	static bool SerializePool(TiXmlElement *container);
	/*------------------------------------------------------
	 * Pre  : 
	 * Post : Serializes all bone objects to XML.
	 *
	 ------------------------------------------------------*/

	static bool UnserializePool(TiXmlElement *container);
	/*------------------------------------------------------
	 * Pre  : 
	 * Post : Unserializes all bone objects from XML.
	 *
	 ------------------------------------------------------*/
#endif


	////////////////////////////////////////////////////////////
	// Public Accessors
	////////////////////////////////////////////////////////////

	hel::Mat44 &GetBindPose() { return mBindPose; }
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : Returns this bone's rest/bind pose transform
	 ------------------------------------------------------*/

	hel::Mat44 &GetInverseBindPose() { return mBindToWorld; }
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : Returns this bone's 'rest to world' transform
	 ------------------------------------------------------*/

	hel::Mat44 &GetWorldPose() { return mTrack.mWorld; }
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : Returns this bone's 'world' transform
	 ------------------------------------------------------*/

	byte GetFlags() { return mFlags; }
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : Get this bone's option flags
	 ------------------------------------------------------*/

	const char *GetName() { return mName; }
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : Get the human readable name of this bone
	 ------------------------------------------------------*/

	index_t GetParent() { return mParent; }
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : Get the UID of this bone's parent bone or 
	 *        INDEX_INVALID if it has no parent
	 ------------------------------------------------------*/

	uint32 GetTrackCount() { return mTrackCount; }
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : 
	 ------------------------------------------------------*/

	BoneTrack &GetTrack(uint32 track) { return mTrack; }
	/*------------------------------------------------------
	 * Pre  : 
	 * Post : Only supporting one 'range/anim' in test: F(track) <- F(0)
	 ------------------------------------------------------*/

	bool Serialize(SystemIO::TextFileWriter &w);
	/*------------------------------------------------------
	 * Pre  :  
	 * Post : 
	 ------------------------------------------------------*/

	mstl::String mMetaData;          /* Metadata for bone */

	byte mFlags;                     /* Options bitmap */

	char mName[64];                  /* Human readable identifier */



private:

	////////////////////////////////////////////////////////////
	// Private Accessors
	////////////////////////////////////////////////////////////


	////////////////////////////////////////////////////////////
	// Private Mutators
	////////////////////////////////////////////////////////////

	mstl::Vector<freyja::Bone*> mBones;             /* */

	mstl::Vector<freyja::Camera*> mCamera;          /* */

	mstl::Vector<freyja::Light*> mLights;           /* */

	mstl::Vector<freyja::Mesh*> mMeshes;            /* */

	mstl::Vector<freyja::Metadata*> mMetadata;      /* */

	mstl::Vector<freyja::Renderable*> mRenderables; /* */

	mstl::Vector<freyja::Skeleton*> mSkeletons;     /* */

	// Break these out into a level above scene, and use instances in scenes.

	mstl::Vector<freyja::Material*> mMaterials;     /* */

	mstl::Vector<freyja::Texture*> mTextures;       /* */
};


} // namespace freyja

#endif // GUARD__FREYJA_BONE_H_

